import math

import astropy.units as u
from astropy.coordinates import SpectralCoord
import numpy as np
from bqplot import LinearScale
from specreduce.tracing import FlatTrace
from specreduce.utils import measure_cross_dispersion_profile
from traitlets import Bool, Float, Integer, List, Unicode, observe

from jdaviz.configs.specviz.plugins.viewers import Spectrum1DViewer
from jdaviz.core.events import GlobalDisplayUnitChanged
from jdaviz.core.marks import PluginLine, PluginMarkCollection, PluginScatter
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (DatasetSelect, PluginTemplateMixin,
                                        PlotMixin)
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               flux_conversion_general)
from jdaviz.core.user_api import PluginUserApi

__all__ = ['CrossDispersionProfile']


@tray_registry('cross-dispersion-profile', label="Cross Dispersion Profile",
               category="data:analysis")
class CrossDispersionProfile(PluginTemplateMixin, PlotMixin):
    """
    The Cross Dispersion Profile plugin allows for visualizaion of the
    cross-dispersion profile of 2d spectra, at a specified wavelength / pixel
    and window.

    The following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :meth:`measure_cross_dispersion_profile`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
        Dataset used to calculate and plot cross-dispersion profile.
    * ``pixel``
        Pixel on spectral axis used to compute and plot profile.
    * ``y_pixel``
        Center of profile on cross-dispersion axis.
    * ``use_full_width``
        If true, full cross-dispersion axis will be used to compute the profile.
    * ``width``
        If use_full_with is False, this value will set the size of the window
        for the profile on the cross-dispersion axis, centered at y_pixel.
    * ``profile``
        Computed profile, as a Quantity array.

    """

    template_file = __file__, "cross_dispersion_profile.vue"

    uses_active_status = Bool(True).tag(sync=True)

    dataset_items = List().tag(sync=True)
    dataset_selected = Unicode().tag(sync=True)

    # pixel on cross dispersion axis where profile will be centered. a FlatTrace
    # at y_pixel will be created to measure the profile.
    y_pixel = Integer().tag(sync=True)

    # pixel on spectral axis to measure profile
    pixel = Integer().tag(sync=True)
    wav = Float(allow_none=True).tag(sync=True)  # corresponding wavelength, if available

    # set maximum values for slider limits
    max_pix = Integer().tag(sync=True)
    max_y_pix = Integer().tag(sync=True)

    # traitlets for size of window in cross-dispersion axis. If 'use_full_width'
    # is True, then the full cross dispersion axis around y_pixel will be used.
    # If False, then 'width' will be used.
    use_full_width = Bool(True).tag(sync=True)
    width = Integer().tag(sync=True)

    # app-wide flux display unit. 'profile' will always be in this unit
    flux_display_unit = Unicode("").tag(sync=True)

    # app-wide unit for spectral axis, for plot title
    sa_display_unit = Unicode("").tag(sync=True)

    # to avoid calculating profile and updating plot when profile parameters
    # are being set for the first time when new data selected
    setting_defaults = Bool().tag(sync=True)

    plot_available = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._marks = {}

        # description displayed under plugin title in tray
        self._plugin_description = 'Visualize cross-dispersion profile.'

        self.dataset = DatasetSelect(self,
                                     'dataset_items',
                                     'dataset_selected',
                                     filters=['layer_in_spectrum_2d_viewer',
                                              'not_trace'])

        self.hub.subscribe(self, GlobalDisplayUnitChanged,
                           handler=self._on_display_units_changed)

        # attribute to access computed profile, will be a quantity array
        self._profile = None

        # override default plot styling
        self.plot.figure.fig_margin = {'top': 60, 'bottom': 60, 'left': 65,
                                       'right': 15}
        self.plot.viewer.axis_y.tick_format = '0.1e'
        self.plot.viewer.axis_y.label_offset = '50px'

        if self.config == "deconfigged":
            self.observe_traitlets_for_relevancy(traitlets_to_observe=['dataset_items'])

    @property
    def user_api(self):
        expose = ('dataset', 'pixel', 'y_pixel', 'use_full_width', 'width',
                  'profile', 'measure_cross_dispersion_profile')
        return PluginUserApi(self, expose=expose)

    @observe("dataset_items")
    def _set_disabled(self, event={}):
        if self.dataset_items == []:
            self.disabled_msg = ('Cross Dispersion Profiles are unavailable when no data'
                                 ' is loaded in the 2D viewer')
        else:
            self.disabled_msg = ''

    @observe("dataset_selected")
    def _set_defaults(self, event={}):
        """
        When a dataset is selected, re-calculate the default values for pixel,
        y_pixel, width, and the slider limits for selecting row/column where
        the profile will be measured.
        """
        # self.dataset might not exist when app is setting itself up.
        if hasattr(self, "dataset") and self.dataset.selected_obj is not None:
            # to avoid entering methods that observe any of these traitlets
            # while they're being set for the first time
            self.setting_defaults = True

            try:
                data = self.dataset.selected_obj
                # default value for 'y_pixel' is middle of cross dispersion axis
                self.y_pixel = math.floor(data.shape[0] / 2)
                # default value for 'pixel' is middle of spectral axis
                self.pixel = math.floor(data.shape[1] / 2)
                # slider limits
                self.max_y_pix = data.shape[0]
                self.max_pix = data.shape[1]
                # default use_full_width=True
                self.use_full_width = True
                # set appropriate default 'width' if use_full_width=False
                self.width = data.shape[0]
            finally:
                # finally, trigger to update plot/profile/marks
                self.setting_defaults = False

    @observe('pixel', 'sa_display_unit')
    def _pixel_to_wav(self, event={}):
        """
        Calculate the corresponding wavelength for ``pixel``, if wcs is present,
        when ``pixel`` is changed (or new dataset selected, in case the previous one
        had a wcs and the new one doesn't or vice versa).
        """
        data = self.dataset.selected_obj
        if data is not None:
            if hasattr(data, 'wcs') and self.sa_display_unit != '':
                wcs = self.dataset.selected_obj.wcs
                # wcs / gwcs don't necessarily have ndim attribute, so try
                # to detect 2d/1d wcs with try / except
                try:  # dataset selected wcs is 1d
                    wav = wcs.pixel_to_world(self.pixel)
                except ValueError:  # dataset selected wcs is 2d
                    wav = [c for c in wcs.pixel_to_world(self.pixel, 0) if isinstance(c, SpectralCoord)]  # noqa
                    if len(wav):
                        wav = wav[0]
                    else:
                        if data.spectral_axis_index == 0:
                            wav = wcs.pixel_to_world(0, self.pixel)[0]
                        else:
                            # It's 2D, so this is the only option
                            wav = wcs.pixel_to_world(self.pixel, 0)[0]

                self.wav = wav.to(u.Unit(self.sa_display_unit), u.spectral()).value
            else:
                self.wav = None

    def _on_display_units_changed(self, event={}):
        """
        On flux display unit change from Unit Conversion plugin, re-compute
        profile in new unit and update plot.

        Note: re-measure profile in native data units rather than converting
        currently computed profile so repeated conversions don't accumulate
        precision errors.
        """
        if event.axis == 'flux':
            if self.flux_display_unit != event.unit:
                self.flux_display_unit = event.unit.to_string()

        if event.axis == 'spectral':
            if self.sa_display_unit != event.unit:
                self.sa_display_unit = event.unit.to_string()

    @property
    def profile(self):
        return self._profile

    @property
    def marks_viewers2d(self):
        return self.dataset.viewers_with_selected_visible

    @property
    def marks_viewers1d(self):
        return self.app.get_viewers_of_cls(Spectrum1DViewer)

    @property
    def marks(self):
        """
        Access the marks created by this plugin in the spectrum-2d-viewer.
        """
        if self._marks:
            return self._marks

        if not self._tray_instance:
            return {}

        self._marks = {'2d': {'pix': PluginMarkCollection(PluginLine,
                                                          visible=self.is_active,
                                                          line_style='solid'),
                              'y_pix': PluginMarkCollection(PluginScatter,
                                                            marker='diamond',
                                                            stroke_width=1)},
                       '1d': {'pix': PluginMarkCollection(PluginLine,
                                                          x=[0, 0],
                                                          y=[0, 1],
                                                          scales={'y': LinearScale(min=0,
                                                                                   max=1)},
                                                          visible=self.is_active,
                                                          line_style='solid')}}

        return self._marks

    @observe('dataset_selected', 'is_active', 'pixel', 'y_pixel', 'width',
             'use_full_width', 'setting_defaults')  # noqa
    def _pixel_selected_mark(self, event={}):
        """
        Update drawn marks (synced vertical lines in 2d and 1d spectrum viewers,
        scatter mark to mark center of profile on y axis) for current selected
        pixel, when any relevant parameter is changed or plugin is made active.
        """
        if self.setting_defaults:
            return

        data = self.dataset.selected_obj
        if data is not None:

            if self.use_full_width is True:
                ymax = data.shape[0]
                ymin = 0
            else:
                ymax = self.y_pixel + int(self.width/2)
                ymin = self.y_pixel - int(self.width/2)

            self.marks['2d']['pix'].update_xy(np.full(data.shape[1], self.pixel),
                                              range(ymin, ymax+1),
                                              viewers=self.marks_viewers2d)
            self.marks['2d']['pix'].visible = self.is_active
            self.marks['2d']['y_pix'].update_xy((self.pixel, self.pixel),
                                                (self.y_pixel, self.y_pixel),
                                                viewers=self.marks_viewers2d)
            self.marks['2d']['y_pix'].visible = self.is_active

            # plot line in 1d viewer when possible, unit conversion is handled
            # inside of Marks so we don't need to convert the limits here
            if self.wav is not None and self.sa_display_unit != '':
                self.marks['1d']['pix'].update_xy((self.wav, self.wav),
                                                  (0, 1),
                                                  viewers=self.marks_viewers1d)
                self.marks['1d']['pix'].visible = self.is_active

    @observe('is_active', 'pixel', 'y_pixel', 'width', 'use_full_width',
             'flux_display_unit', 'setting_defaults')
    def _measure_cross_dispersion_profile(self, event={}):

        if self.setting_defaults:
            return

        self.measure_cross_dispersion_profile(update_plot=self.is_active)

    def measure_cross_dispersion_profile(self, update_plot=True):
        """
        Measure the cross-dispersion profile and update plugin plot.

        Calculates the cross-dispersion profile for the currently
        selected dataset at column ``pixel``. If ``use_full_width`` is True,
        the profile is computed over the entire detector width, otherwise,
        a user-defined ``width`` and center ``y_pixel`` are used. The profile
        is returned and plotted in the app-wide flux display unit, as set in
        the Unit Conversion plugin. If update_plot is True, the plugin plot
        will be updated with the computed profile.

        Parameters
        ----------
        update_plot : bool, optional
            If True, update the plugin plot with the profile (default True).

        Returns
        -------
        profile : array-like
            The computed cross-dispersion profile.

        """

        data = self.dataset.selected_obj
        if data is None:
            return

        if self.use_full_width:
            width = None
        else:
            width = self.width

        # create a FlatTrace at y_pixel
        trace = FlatTrace(data, self.y_pixel)

        profile = measure_cross_dispersion_profile(data,
                                                   trace=trace,
                                                   crossdisp_axis=0,
                                                   width=width,
                                                   pixel=self.pixel,
                                                   pixel_range=None,
                                                   align_along_trace=False)

        # convert profile, which was computed in data units, to display unit
        if self.sa_display_unit != '':
            if self.wav is not None:
                wav = self.wav * u.Unit(self.sa_display_unit)
            else:
                wav = None
            eqv = all_flux_unit_conversion_equivs(data.meta.get('PIXAR_SR', 1.0),
                                                  wav)
            profile = flux_conversion_general(profile.value, profile.unit,
                                              self.flux_display_unit, eqv)

        self._profile = profile

        if update_plot:
            self.update_plot()

    @observe('sa_display_unit')
    def update_plot(self, event={}):
        """Update plugin plot with self.profile."""

        if self.profile is None:
            return

        x = np.arange(len(self.profile))

        if not self.use_full_width:
            # translate x-axis of plot to image y-axis coordinates so plot
            # is centered on y_pixel
            x += int(self.y_pixel - (self.width / 2))

        self.plot._update_data('profile', x=x, y=self.profile, reset_lims=True)
        self.plot.update_style('profile', line_visible=True, color='gray',
                               size=32)

        title = f'Cross dispersion profile for pixel {self.pixel}'

        # include wavelength in plot title, if available
        if self.wav is not None:
            title += f' ({round(self.wav, 3)} {self.sa_display_unit})'
        self.plot.figure.title = title

        self.plot.figure.axes[0].label = 'pixel'
        self.plot.figure.axes[1].label = f'Value ({self.flux_display_unit})'

        self.plot_available = True



from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin

__all__ = ["VoPlugin"]
vo_plugin_label = "Virtual Observatory"


@tray_registry("VoPlugin", label=vo_plugin_label)
class VoPlugin(PluginTemplateMixin):
    """Plugin to query the Virtual Observatory and load data into Imviz"""

    template_file = __file__, "vo_plugin.vue"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = (
            "Download data products from VO-registered telescopes and missions."
        )



"""
Test the launcher module functionality.
"""
from ipywidgets import widgets
import os
import pytest
from unittest.mock import Mock, patch

from astropy.io.registry import IORegistryError
from astropy.nddata import CCDData
from astropy.wcs import WCS
import numpy as np
from pathlib import Path

from jdaviz import Imviz
from jdaviz.cli import (DEFAULT_VERBOSITY,
                        DEFAULT_HISTORY_VERBOSITY,
                        ALL_JDAVIZ_CONFIGS)
from jdaviz.core.launcher import (open as jdaviz_open,
                                  _launch_config_with_data,
                                  Launcher,
                                  show_launcher,
                                  STATUS_HINTS)


@pytest.fixture(scope='module')
def _shared_ccd(tmp_path_factory):
    """
    Create a single FITS file once per module and return the CCD and
    filepath.
    """
    test_dir = tmp_path_factory.mktemp('data')
    test_file = test_dir / 'test_image.fits'
    data = np.ones((10, 10))
    wcs = WCS(naxis=2)
    ccd = CCDData(data=data, unit='adu', wcs=wcs)
    ccd.write(test_file)
    return ccd, str(test_file)


@pytest.fixture(autouse=True, scope='module')
def _module_setup(request, _shared_ccd):
    """
    Module-scoped autouse fixture that attaches the shared CCD and
    filepath to all test classes in the module so tests can use
    ``self.ccd`` and ``self.test_file`` without rewriting the file.
    """
    ccd, test_file = _shared_ccd

    # Attach as class attributes to every Test* class in the module.
    for name in dir(request.module):
        obj = getattr(request.module, name)
        if isinstance(obj, type) and name.startswith('Test'):
            setattr(obj, 'ccd', ccd)
            setattr(obj, 'test_file', test_file)


def test_status_hints_keys():
    """
    Test that STATUS_HINTS contains expected keys.
    """
    expected_keys = {'idle',
                     'identifying',
                     'invalid path',
                     'id ok',
                     'id failed'}
    assert set(STATUS_HINTS.keys()) == expected_keys


class TestOpenFunction:
    """
    Test the open() function for automatic config detection and
    launching.
    """
    @pytest.mark.parametrize('local_path', ['/some/path', None])
    def test_open_with_local_path_kwarg(self, local_path):
        """
        Test open() with local_path kwarg passed through to
        download_uri_to_path and with multiple compatible helpers.
        """
        # Mock the download to avoid actual network calls.
        with (patch('jdaviz.core.launcher.download_uri_to_path') as mock_download):
            mock_download.return_value = self.test_file
            # mock the identify_helper to force imviz to be used with the ccd data
            with patch('jdaviz.core.launcher.identify_helper') as mock_identify:
                mock_identify.return_value = (['imviz'], self.ccd)

                result = jdaviz_open(self.test_file, show=False, local_path=local_path)

                mock_download.assert_called_once_with(self.test_file,
                                                      cache=True,
                                                      local_path=local_path)
                assert isinstance(result, Imviz)

                # When multiple compatible helpers are found, NotImplementedError is raised.
                mock_identify.return_value = (['imviz', 'cubeviz'], self.ccd)

                msg = 'Multiple helpers provided'
                with pytest.raises(NotImplementedError, match=msg):
                    jdaviz_open(self.test_file, show=False)


class TestLaunchConfigWithData:
    """
    Test the _launch_config_with_data() function.
    """
    @pytest.mark.parametrize(('data', 'verbosity', 'history_verbosity'),
                             [(None, None, None),
                              ('', DEFAULT_VERBOSITY, DEFAULT_HISTORY_VERBOSITY),
                              ('', 'debug', 'info')])
    def test_launch_config_basic(self, imviz_helper, data, verbosity, history_verbosity):
        """
        Test basic launching of a config without data.
        """
        kwargs = dict(data=data, verbosity=verbosity,
                      history_verbosity=history_verbosity, show=False)
        if verbosity is None:
            kwargs.pop('verbosity')
        if history_verbosity is None:
            kwargs.pop('history_verbosity')

        # mock the config class, but we can use the imviz_helper fixture here
        # (as opposed to the test below which also mocks the helper)
        # since we're not actually attempting to load anything.
        with patch('jdaviz.core.launcher.jdaviz_configs') as mock_configs:
            mock_configs.Imviz = Mock(return_value=imviz_helper)

            result = _launch_config_with_data('imviz', **kwargs)

            mock_configs.Imviz.assert_called_once_with(
                verbosity=DEFAULT_VERBOSITY if verbosity is None else verbosity,
                history_verbosity=DEFAULT_HISTORY_VERBOSITY if history_verbosity is None else history_verbosity)  # noqa
            assert result == imviz_helper

    def test_launch_config_with_data_loads_successfully(self):
        """
        Test launching config with data that loads successfully.
        """
        mock_helper = Mock()
        mock_helper.load_data = Mock()
        mock_helper.show = Mock()

        # mock the config class to return our mock helper
        # so that we don't have to worry about the actual loading logic
        # from load_data.
        with patch('jdaviz.core.launcher.jdaviz_configs') as mock_configs:
            mock_configs.Imviz = Mock(return_value=mock_helper)

            result = _launch_config_with_data('imviz',
                                              data=self.ccd,
                                              show=False,
                                              data_label='test')

            mock_helper.load_data.assert_called_once_with(self.ccd, data_label='test')
            mock_helper.show.assert_not_called()
            assert result == mock_helper

    def test_launch_config_with_data_io_errors(self):
        """
        Test that when loading data raises IORegistryError.
        Check with filepath fallback and without.
        """
        mock_helper = Mock()
        mock_helper.load_data = Mock(side_effect=[IORegistryError('Failed'), None])
        mock_helper.show = Mock()

        with patch('jdaviz.core.launcher.jdaviz_configs') as mock_configs:
            mock_configs.Imviz = Mock(return_value=mock_helper)

            _ = _launch_config_with_data('imviz',
                                         data=self.ccd,
                                         filepath=self.test_file)

            assert mock_helper.load_data.call_count == 2
            # Check that show is called because show=True (default)
            mock_helper.show.assert_called_once()
            mock_helper.load_data.assert_any_call(self.ccd)
            mock_helper.load_data.assert_any_call(self.test_file)

        # Now check that without filepath, the error is raised.
        mock_helper.load_data = Mock(side_effect=IORegistryError('Failed'))

        with patch('jdaviz.core.launcher.jdaviz_configs') as mock_configs:
            mock_configs.Imviz = Mock(return_value=mock_helper)

            with pytest.raises(IORegistryError, match='Failed'):
                _launch_config_with_data('imviz', data='some_data', show=False)


class TestLauncherClass:
    """
    Test the Launcher class.
    """
    def test_launcher_init_default(self):
        """
        Test Launcher initialization with default parameters.
        """
        launcher = Launcher()

        assert launcher.configs == ALL_JDAVIZ_CONFIGS
        assert launcher.compatible_configs == ALL_JDAVIZ_CONFIGS
        assert launcher.filepath == ''
        assert launcher.loaded_data is None
        assert launcher.hint == STATUS_HINTS['idle']
        assert launcher.file_browser_visible is False
        assert launcher.file_browser_widget is None
        for config in ALL_JDAVIZ_CONFIGS:
            assert config in launcher.config_icons

    def test_launcher_init_with_custom_init(self):
        """
        Test Launcher initialization with custom config list.
        """
        custom_configs = ['imviz', 'specviz']
        launcher = Launcher(configs=custom_configs)

        assert launcher.configs == custom_configs
        assert launcher.compatible_configs == custom_configs

        launcher = Launcher(height=600)
        assert launcher.height == '600px'

        launcher = Launcher(height='100%')
        assert launcher.height == '100%'

        # Test main_with_launcher property returns main with launcher as child.
        result = launcher.main_with_launcher
        assert result == launcher.main
        assert launcher.main.children == [launcher]

        mock_main = Mock()
        launcher = Launcher(main=mock_main)
        assert launcher.main == mock_main

        # vdocs is set to 'latest' for dev versions.
        with patch('jdaviz.core.launcher.__version__', '1.0.dev'):
            launcher = Launcher()
            assert launcher.vdocs == 'latest'

        with patch('jdaviz.core.launcher.__version__', '1.2.3'):
            launcher = Launcher()
            assert launcher.vdocs == 'v1.2.3'

        # Test that JDAVIZ_START_DIR environment variable is respected.
        test_dir = '/custom/start/dir'
        # Use patch.dict to temporarily set the environment variable.
        with patch.dict(os.environ, {'JDAVIZ_START_DIR': test_dir}):
            launcher = Launcher()
            launcher.vue_open_file_dialog()
            assert Path(launcher.file_browser_dir.value) == Path(os.path.abspath(test_dir))

    def test_launcher_init_with_filepath(self):
        """
        Test Launcher initialization with various filepaths.
        """
        launcher = Launcher()

        launcher.filepath = 'something that does not exist'
        assert launcher.hint == STATUS_HINTS['invalid path']
        assert launcher.compatible_configs == []
        assert launcher.loaded_data is None

        # Then reset and check that we're in an idle state
        launcher.filepath = ''
        assert launcher.hint == STATUS_HINTS['idle']
        assert launcher.compatible_configs == ALL_JDAVIZ_CONFIGS
        assert launcher.loaded_data is None

        with patch('jdaviz.core.launcher.identify_helper') as mock_identify:
            mock_identify.return_value = (['imviz'], self.ccd)

            launcher = Launcher(filepath=self.test_file)

            assert launcher.filepath == self.test_file
            assert launcher.hint == STATUS_HINTS['id ok']
            assert launcher.compatible_configs == ['imviz']
            assert launcher.loaded_data == self.ccd

    def test_filepath_changed_identify(self):
        """
        Test that identification failure shows all configs and uses
        filepath as data. Also test when identify_helper returns empty config list
        with None data - should leave loaded_data as None.
        """
        with patch('jdaviz.core.launcher.identify_helper') as mock_identify:
            mock_identify.side_effect = Exception('Identification failed')

            launcher = Launcher()
            launcher.filepath = self.test_file

            assert launcher.hint == STATUS_HINTS['id failed']
            assert launcher.compatible_configs == ALL_JDAVIZ_CONFIGS
            assert launcher.loaded_data == self.test_file

        with patch('jdaviz.core.launcher.identify_helper') as mock_identify:
            mock_identify.return_value = ([], None)

            launcher = Launcher()
            launcher.filepath = self.test_file

            # When configs is empty but data is None, loaded_data stays None
            assert launcher.loaded_data is None
            assert launcher.compatible_configs == []

    def test_vue_choose_file(self, tmp_path):
        """
        Test vue_choose_file with various scenarios.
        """
        launcher = Launcher()
        launcher.vue_open_file_dialog()
        assert launcher.file_browser_widget is not None

        # No file
        launcher.selected_file.value = ''
        launcher.vue_choose_file()

        assert launcher.filepath == ''

        # Choose with directory (should not update filepath)
        launcher.selected_file.value = 'somedir'
        launcher.file_browser_dir.value = str(tmp_path)
        launcher.file_browser_visible = True

        launcher.vue_choose_file()

        # Should close dialog even if it's a directory
        assert launcher.file_browser_visible is False

        with patch('jdaviz.core.launcher.identify_helper') as mock_identify:
            mock_identify.return_value = (['imviz'], self.ccd)

            test_file_path = Path(self.test_file)
            launcher.file_browser_dir.value = test_file_path.parent
            launcher.selected_file.value = test_file_path.name

            launcher.vue_choose_file()

            assert launcher.file_browser_visible is False
            assert launcher.filepath == self.test_file

    @pytest.mark.parametrize('height', ['600', '100%', '100vh'])
    def test_vue_launch_config(self, deconfigged_helper, height):
        """
        Test vue_launch_config method.
        """
        mock_helper = Mock()
        dcf_app = deconfigged_helper.app
        mock_helper.app = dcf_app
        default_height = '800px'
        dcf_app.state.settings = {'context': {'notebook': {'max_height': default_height}}}
        dcf_app.layout = widgets.Layout(height=height, width="100%")

        with patch('jdaviz.core.launcher.identify_helper') as mock_identify:
            mock_identify.return_value = (['imviz'], self.ccd)

            with patch('jdaviz.core.launcher._launch_config_with_data') as mock_launch:
                mock_launch.return_value = mock_helper

                launcher = Launcher(height=height)
                launcher.filepath = self.test_file

                event = {'config': 'imviz'}
                launcher.vue_launch_config(event)

                mock_launch.assert_called_once_with('imviz',
                                                    self.ccd,
                                                    filepath=self.test_file,
                                                    show=False)

                assert launcher.main.color == 'transparent'
                assert launcher.main.children == [dcf_app]

                if height not in ['100%', '100vh']:
                    assert dcf_app.layout.height == default_height
                    assert launcher.main.height == default_height


class TestShowLauncher:
    """
    Test the show_launcher() function.
    """
    def test_show_launcher_default(self):
        """
        Test show_launcher with default parameters.
        """
        # Patch the Launcher class to avoid full initialization
        # and unintended side effects.
        with patch('jdaviz.core.launcher.Launcher') as mock_launcher_class:
            mock_launcher = Mock()
            mock_launcher.main_with_launcher = Mock()
            mock_launcher_class.return_value = mock_launcher

            # Patch show_widget to avoid display calls.
            with patch('jdaviz.core.launcher.show_widget') as mock_show:
                show_launcher()

                mock_launcher_class.assert_called_once_with(None, ALL_JDAVIZ_CONFIGS, '', '450px')
                mock_show.assert_called_once_with(
                    mock_launcher.main_with_launcher,
                    loc='inline',
                    title=None)

    @pytest.mark.parametrize(('height', 'str_height'),
                             [(800, '800px'),
                              ('100%', '100%'),
                              ('100vh', '100vh')])
    def test_show_launcher_with_custom_params(self, height, str_height):
        """
        Test show_launcher with custom parameters.
        """
        custom_configs = ['imviz', 'specviz']
        custom_filepath = '/path/to/file.fits'

        with patch('jdaviz.core.launcher.Launcher') as mock_launcher_class:
            mock_launcher = Mock()
            mock_launcher.main_with_launcher = Mock()
            mock_launcher_class.return_value = mock_launcher

            with patch('jdaviz.core.launcher.show_widget') as _:
                show_launcher(configs=custom_configs, filepath=custom_filepath, height=height)

                mock_launcher_class.assert_called_once_with(
                    None,
                    custom_configs,
                    custom_filepath,
                    str_height)



import numpy as np
import pytest
from astropy import units as u
from specutils import Spectrum
from jdaviz.core.marks import PluginLine


@pytest.mark.xfail(reason="Currently fails due to UnitConversionError with pixel units")
def test_multiple_2d_spectra(deconfigged_helper, spectrum2d, mos_spectrum2d):
    """
    Test mouseover handling for multiple 2D spectra with different wavelength mappings.

    This test module covers the scenario where multiple 2D spectra are loaded with:
    - WCS-based wavelength mapping (using astropy WCS)
    - Simple spectral axis-based wavelength mapping
    - No wavelength mapping (pixel coordinates only)

    Mouseover previews should appear on all relevant viewers.
    """
    def _mouseover_and_check_line_visible(viewer, label_mouseover, x, y):
        # Simulate mouseover event at a specific position
        label_mouseover._viewer_mouse_event(viewer,
                                            {'event': 'mousemove', 'domain': {'x': x, 'y': y}})
        line_visible = False
        for mark in viewer.figure.marks:
            if isinstance(mark, PluginLine) and mark.visible:
                line_visible = True
                break
        return line_visible

    deconfigged_helper.load(mos_spectrum2d, data_label='2D Spectrum WCS',
                            format='2D Spectrum')

    viewer_1d = deconfigged_helper.viewers['1D Spectrum']._obj.glue_viewer
    viewer_2d = deconfigged_helper.viewers['2D Spectrum']._obj.glue_viewer

    # Get the coordinates info object for mouseover
    label_mouseover = deconfigged_helper._coords_info

    # Test mouseover on 2D spectrum with WCS
    assert _mouseover_and_check_line_visible(viewer_2d, label_mouseover, 5, 5)
    # Test mouseover on 1d spectrum linked to the 2D spectrum
    assert _mouseover_and_check_line_visible(viewer_1d, label_mouseover, 5, 5)

    # Add a second 2D spectrum with simple spectral axis (no WCS)
    deconfigged_helper.load(spectrum2d, data_label='2D Spectrum Spectral Axis',
                            format='2D Spectrum')
    assert _mouseover_and_check_line_visible(viewer_2d, label_mouseover, 5, 5)
    assert _mouseover_and_check_line_visible(viewer_1d, label_mouseover, 5, 5)

    data_no_wcs = np.ones((12, 18)) * 2.0
    data_no_wcs[6] = np.arange(18) * 0.3
    # Create spectrum without spectral_axis or wcs - just pixel coordinates
    spectrum2d_no_wavelength = Spectrum(flux=data_no_wcs*u.Jy, spectral_axis_index=0)

    # TODO: fix failure 'astropy.units.errors.UnitConversionError: 'pix' and 'm' (length)
    #  are not convertible'
    deconfigged_helper.load(spectrum2d_no_wavelength, data_label='2D Spectrum No Wavelength',
                            format='2D Spectrum')
    assert _mouseover_and_check_line_visible(viewer_2d, label_mouseover, 5, 5)
    assert _mouseover_and_check_line_visible(viewer_1d, label_mouseover, 5, 5)



import os
import warnings
from contextlib import contextmanager
from functools import cached_property
from traitlets import Bool, Float, Instance, List, Unicode, observe, default
from ipywidgets import widget_serialization

from glue_jupyter.common.toolbar_vuetify import read_icon
from astropy.coordinates import SkyCoord
from astropy.coordinates.builtin_frames import __all__ as all_astropy_frames
from astropy.table import Table as astropyTable
from astroquery.mast import MastMissions

from jdaviz.core.custom_traitlets import FloatHandleEmpty, IntHandleEmpty
from jdaviz.core.events import (AddDataMessage,
                                RemoveDataMessage,
                                SnackbarMessage,
                                FootprintOverlayClickMessage,
                                LinkUpdatedMessage,
                                ViewerAddedMessage)
from jdaviz.core.marks import RegionOverlay
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        SelectPluginComponent,
                                        Table,
                                        CustomToolbarToggleMixin,
                                        UnitSelectPluginComponent,
                                        ViewerSelect,
                                        with_spinner,
                                        _is_image_viewer)
from jdaviz.core.registries import (loader_resolver_registry,
                                    loader_parser_registry,
                                    loader_importer_registry)
from jdaviz.core.user_api import LoaderUserApi
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.region_translators import is_stcs_string, stcs_string2region
from jdaviz.utils import download_uri_to_path, find_closest_polygon_mark, layer_is_image_data
from glue.core.message import DataCollectionAddMessage, DataCollectionDeleteMessage


__all__ = ['BaseResolver', 'BaseConeSearchResolver', 'find_matching_resolver']


class FormatSelect(SelectPluginComponent):
    """
    Select component for the format field of a resolver.

    Parameters
    ----------
    plugin
        the parent plugin object
    items : str
        the name of the items traitlet defined in ``plugin``
    selected : str
        the name of the selected traitlet defined in ``plugin``
    default_mode : str, optional
        What mode to use when making the default selection.  Valid options: first, default_text,
        empty.
    """
    debug = Bool(False).tag(sync=True)
    _parsers = {}
    _importers = {}

    def __init__(self, plugin, items, selected, default_mode='first'):
        self._invalid_importers = {}
        self._importers = {}
        super().__init__(plugin,
                         items=items,
                         selected=selected,
                         default_mode=default_mode)

    def _is_valid_item(self, item):
        return super()._is_valid_item(item, locals())

    @observe('filters', 'debug')
    def _update_items(self, msg={}):
        if not self.plugin.is_valid:
            self.items = []
            self._apply_default_selection()
            return

        all_formats = []
        self._parsers = {}
        self._dbg_importers = {}
        self._invalid_importers = {}
        self._importers = {}

        # check for valid parser > importer combinations given the current filters
        # and resolver inputs
        try:
            # NOTE: plugin is just because this inherits from SelectPluginComponent,
            # but is actually the resolver.  This calls the implemented __call__ method
            # on the parent resolver.
            parser_input = self.plugin.output
        except Exception as e:
            self.items = []
            self._invalid_importers = f'resolver exception: {e}'
            self._apply_default_selection()
            return

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for parser_name, Parser in loader_parser_registry.members.items():
                this_parser = Parser(self.plugin.app, parser_input)
                self._parsers[parser_name] = this_parser
                try:
                    if this_parser.is_valid:
                        importer_input = this_parser.output
                    else:
                        self._invalid_importers[parser_name] = 'not valid'
                        importer_input = None
                except Exception as e:
                    self._invalid_importers[parser_name] = f'parser exception: {e}'
                    importer_input = None

                if importer_input is None:
                    self._invalid_importers.setdefault(parser_name, 'importer_input is None')
                    this_parser._cleanup()
                    continue
                for importer_name, Importer in loader_importer_registry.members.items():
                    label = f"{parser_name} > {importer_name}"
                    if self.plugin._restrict_to_formats is not None and \
                            importer_name not in self.plugin._restrict_to_formats:
                        self._invalid_importers[label] = 'not matching format restriction'  # noqa
                        continue
                    try:
                        this_importer = Importer(app=self.plugin.app,
                                                 resolver=self.plugin,
                                                 parser=this_parser,
                                                 input=importer_input)
                    except Exception as e:  # nosec
                        self._invalid_importers[label] = f'importer exception: {e}'
                        continue
                    if self.debug:
                        self._dbg_importers[label] = this_importer
                    if (self.plugin._restrict_to_target is not None and
                            this_importer.target.get('label') != self.plugin._restrict_to_target):
                        # skip importers that do not match the target
                        self._invalid_importers[label] = 'not matching target'
                        continue
                    if this_importer.is_valid:
                        if self._is_valid_item(this_importer):
                            item = {'label': importer_name,
                                    'parser': parser_name,
                                    'importer': importer_name,
                                    'target': this_importer.target}
                            parser_pref = this_importer.parser_preference
                            if importer_name not in self._importers:
                                all_formats.append(item)
                                self._importers[importer_name] = this_importer
                            elif not len(parser_pref) or parser_name not in parser_pref:
                                # default to the previous (or first) found match
                                continue
                            else:
                                # then there was already a match from an earlier parser.  Compare
                                # to see which has preference and replace if necessary.
                                item_importers = [i['importer'] for i in all_formats]
                                item_index = item_importers.index(importer_name)
                                prev_parser = all_formats[item_index]['parser']
                                parser_pref = this_importer.parser_preference
                                if (prev_parser not in parser_pref or
                                        parser_pref.index(prev_parser) > parser_pref.index(parser_name)):  # noqa
                                    # this parser has preference over the previous one
                                    all_formats[item_index] = item
                                    self._importers[importer_name] = this_importer
                                else:
                                    # this previous parser has preference over this one
                                    continue

                        else:
                            # we'll store the importer even if it isn't valid according to the
                            # filters so that they can be used when compiling the list of
                            # target filters
                            self._importers[importer_name] = this_importer
                    else:
                        self._invalid_importers[label] = 'not valid'

        self.items = all_formats
        self._apply_default_selection()


class TargetSelect(SelectPluginComponent):
    """
    Select component for the target (filter on format) of a resolver.

    Parameters
    ----------
    plugin
        the parent plugin object
    items : str
        the name of the items traitlet defined in ``plugin``
    selected : str
        the name of the selected traitlet defined in ``plugin``
    default_mode : str, optional
        What mode to use when making the default selection.  Valid options: first, default_text,
        empty.
    """

    def __init__(self, plugin, items, selected, default_mode='first'):
        self._importers = {}
        super().__init__(plugin,
                         items=items,
                         selected=selected,
                         default_mode=default_mode)

    def _is_valid_item(self, item):
        return super()._is_valid_item(item, locals())

    def set_filter_target_in(self, targets):
        def filter_factory(targets):
            def filter_target(item):
                return item['label'] in targets
            return filter_target
        self.filters = [filter_factory(targets)]

    @observe('filters')
    def _update_items(self, msg={}):
        if not self.plugin.is_valid:
            self.items = []
            self._apply_default_selection()
            return

        # loop through choices on format and compile list of targets
        # note that the selection of a target may affect the available formats
        # so we want to store all importers in the target select even if they are not valid there
        # and use that list when compiling list of valid targets
        all_targets = []
        for importer in self.plugin.format._importers.values():
            target = importer.target
            if target not in all_targets:
                all_targets.append(target)

        self.items = [{'label': 'Any'}] + [item for item in all_targets if self._is_valid_item(item)]  # noqa
        self._apply_default_selection()


class BaseResolver(PluginTemplateMixin, CustomToolbarToggleMixin):
    _defer_resolver_input_updated = False  # noqa: only use via defer_resolver_input_updated context manager
    default_input = None
    default_input_cast = None
    requires_api_support = False

    spinner = Unicode("").tag(sync=True)

    parsed_input_is_empty = Bool(True).tag(sync=True)

    # whether the current output could be interpretted as a list of data products
    parsed_input_is_query = Bool(False).tag(sync=True)
    # if parsed_input_is_query is True, whether to treat it as such
    # or pass it on directly to the parsers/importers
    treat_table_as_query = Bool(True).tag(sync=True)

    observation_table = Instance(Table).tag(sync=True, **widget_serialization)
    observation_table_populated = Bool(False).tag(sync=True)
    file_table = Instance(Table).tag(sync=True, **widget_serialization)
    file_table_populated = Bool(False).tag(sync=True)

    # options to download selected item in products list
    file_cache = Bool(True).tag(sync=True)
    file_local_path = Unicode("./").tag(sync=True)
    file_timeout = FloatHandleEmpty(10).tag(sync=True)

    importer_widget = Unicode().tag(sync=True)

    # Set remote server options based on the app configuration
    # read-only: change via app.state.settings['server_is_remote']
    server_is_remote = Bool(False).tag(sync=True)

    format_items = List().tag(sync=True)
    format_selected = Unicode().tag(sync=True)
    valid_import_formats = Unicode().tag(sync=True)

    target_items = List().tag(sync=True)
    target_selected = Unicode().tag(sync=True)

    is_wcs_linked = Bool(False).tag(sync=True)
    image_data_loaded = Bool(False).tag(sync=True)
    footprint_select_icon = Unicode(read_icon(os.path.join(
        ICON_DIR, 'footprint_select.svg'), 'svg+xml')).tag(sync=True)

    def __init__(self, *args, **kwargs):
        self.set_active_loader_callback = kwargs.pop('set_active_loader_callback', None)
        self.open_callback = kwargs.pop('open_callback', None)
        self.close_callback = kwargs.pop('close_callback', None)
        self._restrict_to_target = kwargs.pop('restrict_to_target', None)

        super().__init__(*args, **kwargs)

        self.observation_table.enable_clear = False
        self.observation_table.show_if_empty = False
        self.observation_table.show_rowselect = True
        self.observation_table.item_key = "Dataset"
        self.observation_table.multiselect = True
        self.observation_table._selected_rows_changed_callback = self.on_observation_select_changed

        self.file_table.enable_clear = False
        self.file_table.show_if_empty = False
        self.file_table.show_rowselect = True
        self.file_table.item_key = "url"
        self.file_table.multiselect = False
        self.file_table._selected_rows_changed_callback = self.on_file_select_changed

        # Setup footprint selection
        if self.app is not None:
            self.is_wcs_linked = getattr(self.app, '_align_by', None) == 'wcs'
            self.app.hub.subscribe(self, FootprintOverlayClickMessage,
                                   handler=self._on_region_select)
            self.image_data_loaded = any(layer_is_image_data(data)
                                         for data in self.app.data_collection)
            self.app.hub.subscribe(self, DataCollectionAddMessage,
                                   handler=self._on_collection_data_added)
            self.app.hub.subscribe(self, DataCollectionDeleteMessage,
                                   handler=self._on_data_removed)
            self.app.hub.subscribe(self, LinkUpdatedMessage,
                                   handler=self._on_link_type_updated)
            self.app.hub.subscribe(self, ViewerAddedMessage,
                                   handler=self._on_viewer_added)
            self.app.hub.subscribe(self, AddDataMessage,
                                   handler=self._on_viewer_data_added)

        def custom_toolbar(viewer):
            if (self.parsed_input_is_query and self.treat_table_as_query and
                    's_region' in self.observation_table.headers_avail):
                return viewer.toolbar._original_tools_nested[:3] + [['jdaviz:selectregion']], 'jdaviz:selectregion'  # noqa: E501
            return None, None

        self.custom_toolbar.callable = custom_toolbar
        self.custom_toolbar.name = "Footprint Selection"

        # subclasses should call self._resolver_input_updated on any change
        # to user-inputs that might affect the available formats
        self.format = FormatSelect(self,
                                   items='format_items',
                                   selected='format_selected')
        self._restrict_to_formats = kwargs.get('format', None)
        if isinstance(self._restrict_to_formats, str):
            self._restrict_to_formats = [self._restrict_to_formats]

        self.target = TargetSelect(self,
                                   items='target_items',
                                   selected='target_selected')

        # Ensure traitlet and app state are in sync at init
        self.server_is_remote = self.app.state.settings.get('server_is_remote',
                                                            self.server_is_remote)

        # Set up bidirectional synchronization
        # Listen for changes to app.state.settings and update traitlet
        self.app.state.add_callback('settings', self._on_app_settings_changed)

    def vue_link_by_wcs(self, *args):
        plg = self.app._jdaviz_helper.plugins.get('Orientation', None)
        if plg is not None:
            plg.align_by = 'WCS'
        self.is_wcs_linked = True

    @default('observation_table')
    def _default_observation_table(self):
        return Table(self,
                     name='observation_table',
                     title='Observations')

    @default('file_table')
    def _default_file_table(self):
        return Table(self,
                     name='file_table',
                     title='Files')

    def _on_app_settings_changed(self, new_settings_dict):
        """
        Update traitlet when app state settings change.

        Parameters
        ----------
        new_settings_dict : dict
            The new settings dictionary from the app state.
        """
        self.server_is_remote = new_settings_dict.get('server_is_remote', False)

    def _on_collection_data_added(self, msg):
        self.image_data_loaded = any(layer_is_image_data(data) for data in self.app.data_collection)

    def _on_viewer_data_added(self, msg):
        """
        If footprint selection is enabled and this is an image viewer
        with valid WCS, add footprints to this viewer.
        """

        if not self.custom_toolbar_enabled:
            return
        if not (self.parsed_input_is_query and self.treat_table_as_query):
            return
        if 's_region' not in self.observation_table.headers_avail:
            return
        # Get the viewer from the message
        if not hasattr(msg, 'viewer_id') or msg.viewer_id is None:
            return

        viewer = self.app.get_viewer_by_id(msg.viewer_id)
        if viewer is None:
            return
        # Check if it's an image viewer
        if not _is_image_viewer(viewer):
            return
        # Does this viewer already have footprints? If so, don't re-add.
        existing_labels = [
            mark.label for mark in viewer.figure.marks
            if isinstance(mark, RegionOverlay)
        ]
        if existing_labels:
            return
        # Check for valid WCS
        if (not hasattr(viewer.state, "reference_data") or
            viewer.state.reference_data is None or
            not hasattr(viewer.state.reference_data, "coords") or
                viewer.state.reference_data.coords is None):
            # No WCS yet; nothing to do
            return
        self._add_footprints_to_viewer(viewer)

    def _on_data_removed(self, msg):
        self.image_data_loaded = any(layer_is_image_data(data) for data in self.app.data_collection)

    def _on_link_type_updated(self, msg=None):
        self.is_wcs_linked = getattr(self.app, '_align_by', None) == 'wcs'

    @contextmanager
    def defer_resolver_input_updated(self):
        """
        Context manager to delay updating format items until multiple traitlets have been set.
        """
        self._defer_resolver_input_updated = True
        try:
            yield
        finally:
            self._defer_resolver_input_updated = False
            self._resolver_input_updated()

    @classmethod
    def from_input(cls, app, inp, format=None, **kwargs):
        self = cls(app=app, format=format)
        if self.default_input is None:
            raise NotImplementedError("Resolver subclass must implement default_input")  # noqa pragma: nocover
        with self.defer_resolver_input_updated():
            setattr(self, self.default_input,
                    self.default_input_cast(inp) if self.default_input_cast else inp)
            user_api = self.user_api
            for k, v in kwargs.items():
                if hasattr(user_api, k):
                    setattr(user_api, k, v)
        return self

    @property
    def is_valid(self):
        # override by subclass
        return False  # pragma: nocover

    @property
    def input(self):
        if self.default_input is None:
            raise NotImplementedError("Resolver subclass must implement default_input")
        return getattr(self, self.default_input)

    def parse_input(self):
        # override by subclass - this should return something that is either interpreted as
        # a products list OR something that can be passed to at least one parser
        raise NotImplementedError("Resolver subclass must implement parse_input")  # pragma: nocover

    @cached_property
    def parsed_input(self):
        return self.parse_input()

    def _parsed_input_to_table(self, parsed_input):
        if (isinstance(parsed_input, str)
                and os.path.exists(parsed_input) and os.path.isfile(parsed_input)):
            # try to read into a table which could be a products list
            try:
                parsed_input = astropyTable.read(parsed_input)
            except Exception:  # nosec
                return None
        if isinstance(parsed_input, astropyTable):
            return parsed_input
        return None

    def _parsed_input_to_observation_table(self, parsed_input_table):
        if 'Dataset' in parsed_input_table.colnames:
            return parsed_input_table
        for map_to_ds in ('fileSetName', 'sci_data_set_name'):
            if map_to_ds in parsed_input_table.colnames:
                parsed_input_table.rename_column(map_to_ds, 'Dataset')
                return parsed_input_table
        return None

    def _parsed_input_to_file_table(self, parsed_input_table):
        if 'url' in parsed_input_table.colnames:
            return parsed_input_table
        for map_to_url in ('URL', 'uri', 'URI', 'dataURI', 'download', 'Filename'):
            if map_to_url in parsed_input_table.colnames:
                parsed_input_table.rename_column(map_to_url, 'url')
                return parsed_input_table
        return None

    def _cleanup(self):
        # clear the existing cache and close any open file references
        # from referenced parsers/importers
        for parser in self.format._parsers.values():
            parser._cleanup()
        for importer in self.format._importers.values():
            if hasattr(importer, '_cleanup'):
                importer._cleanup()
        self._clear_cache('parsed_input', 'output')

    @observe('parsed_input_is_query', 'treat_table_as_query')
    @with_spinner('spinner', 'parsing input...')
    def _resolver_input_updated(self, msg={}):
        if self._defer_resolver_input_updated:
            return
        if msg.get('name') == 'treat_table_as_query':
            # the input itself has remain unchanged, but how that
            # is mapped to output has
            self._clear_cache('output')
        else:
            self._cleanup()

        try:
            parsed_input = self.parsed_input  # calls self.parse_input() on the subclass and caches
        except Exception:  # nosec
            self.parsed_input_is_empty = True
            self.parsed_input_is_query = False
            self.observation_table_populated = False
            self.file_table_populated = False
            self.observation_table._clear_table()
            self.file_table._clear_table()
            self._update_format_items()
            return

        if parsed_input is None or getattr(parsed_input, 'len', lambda: 1)() == 0:
            self.parsed_input_is_empty = True
            self.parsed_input_is_query = False
            self.observation_table_populated = False
            self.file_table_populated = False
            self.observation_table._clear_table()
            self.file_table._clear_table()
            self._update_format_items()
            return

        # first attempt to parse the input as a table
        parsed_input_table = self._parsed_input_to_table(parsed_input)
        # if the input could be parsed as a table, try to interpret it as
        # either an observation table or file table.  parsed_input_table
        # will be None if it could not be parsed as a table.
        if parsed_input_table is not None:
            file_table = self._parsed_input_to_file_table(parsed_input_table)
            if file_table is not None:
                self.observation_table._clear_table()
                self.file_table._clear_table()

                for row in file_table:
                    self.file_table.add_item(row)
                self.parsed_input_is_query = True
                self.observation_table_populated = False
                self.file_table_populated = True
                return

            observation_table = self._parsed_input_to_observation_table(parsed_input_table)
            if observation_table is not None:
                self.observation_table._clear_table()
                self.file_table._clear_table()

                for row in observation_table:
                    self.observation_table.add_item(row)
                self.parsed_input_is_query = True
                self.observation_table_populated = True
                self.file_table_populated = False
                return

        self.parsed_input_is_empty = False
        self.parsed_input_is_query = False
        self.observation_table_populated = False
        self.file_table_populated = False
        self._update_format_items()

    @cached_property
    def missions_query(self):
        return MastMissions()

    @with_spinner('spinner', 'fetching product list...')
    def _get_product_list(self, mission, dataset):
        self.missions_query.mission = mission
        return self.missions_query.get_product_list(dataset)

    @staticmethod
    def guess_mission(dataset):
        if dataset.startswith('jw'):
            return 'jwst'
        elif dataset.startswith('r'):
            return 'roman'
        else:
            return 'hst'

    def on_observation_select_changed(self, _=None):
        # Sync footprint selection across all viewers (only if footprints are displayed)
        if self.custom_toolbar_enabled and self.parsed_input_is_query and self.treat_table_as_query:
            self._sync_footprint_selection_to_viewers()
        # Fetch products if rows are selected
        if len(self.observation_table.selected_rows) == 0:
            self.app.hub.broadcast(SnackbarMessage("No observation currently selected",
                                                   sender=self, color="warning"))
        else:
            datasets = [row['Dataset'] for row in self.observation_table.selected_rows]
            results = self._get_product_list(self.guess_mission(datasets[0]), datasets)
            file_table = self._parsed_input_to_file_table(results)
            if file_table is not None:
                self.file_table._clear_table()
                for row in file_table:
                    self.file_table.add_item(row)
                self.file_table_populated = True
            else:
                self.app.hub.broadcast(SnackbarMessage(f"No products found for {datasets}",
                                                       sender=self, color="error"))
                self.file_table_populated = False

    def _get_image_viewers(self):
        return [viewer for viewer in self.app._viewer_store.values()
                if _is_image_viewer(viewer)]

    def toggle_custom_toolbar(self):
        """Override to control footprint display when toolbar is toggled."""
        if not self.custom_toolbar_enabled:
            if self.parsed_input_is_query and self.treat_table_as_query:
                self._display_observation_footprints()
            super().toggle_custom_toolbar()
        else:
            super().toggle_custom_toolbar()
            if self.parsed_input_is_query and self.treat_table_as_query:
                self._remove_observation_footprints()

    @observe('custom_toolbar_enabled')
    def _on_custom_toolbar_enabled_changed(self, change={}):
        if not change.get('new', self.custom_toolbar_enabled):
            # Toolbar was disabled - remove footprints from all viewers
            if self.parsed_input_is_query and self.treat_table_as_query:
                self._remove_observation_footprints()

    def _on_region_select(self, msg):
        """Handle footprint click events."""
        if not (self.parsed_input_is_query and self.treat_table_as_query):
            return

        click_viewer = msg.sender.viewer if hasattr(msg.sender, 'viewer') else None
        if click_viewer is None:
            return
        region_marks = [
            mark for mark in click_viewer.figure.marks
            if isinstance(mark, RegionOverlay)
        ]

        click_x, click_y = msg.x, msg.y
        selected_idx = find_closest_polygon_mark(click_x, click_y, region_marks)

        if selected_idx is not None:
            currently_selected = set()
            for row in self.observation_table.selected_rows:
                idx = self.observation_table.items.index(row)
                currently_selected.add(idx)

            # Toggle selection
            if selected_idx in currently_selected:
                currently_selected.discard(selected_idx)
            else:
                currently_selected.add(selected_idx)

            # Update the table selection
            if currently_selected:
                self.observation_table.select_rows(sorted(list(currently_selected)))
            else:
                # Clear selection
                self.observation_table.selected_rows = []

    def _display_observation_footprints(self):
        if 's_region' not in self.observation_table.headers_avail:
            return

        image_viewers = self._get_image_viewers()
        if not image_viewers:
            return
        regions = []
        region_labels = []
        for idx, row in enumerate(self.observation_table.items):
            s_region_str = row.get('s_region', '')

            if not s_region_str or (isinstance(s_region_str, str) and s_region_str.strip() == ''):
                continue
            if not is_stcs_string(s_region_str):
                continue
            region = stcs_string2region(s_region_str)
            regions.append(region)
            region_labels.append(idx)

        if not regions:
            return

        for viewer in image_viewers:
            viewer._add_region_overlay(
                region=regions,
                region_label=region_labels,
                selected=False,
                fill_opacities=[0]
            )

        self._sync_footprint_selection_to_viewers()

    def _remove_observation_footprints(self):
        """Remove all observation footprints from all image viewers."""
        image_viewers = self._get_image_viewers()
        if not image_viewers:
            return

        # Get all labels (row indices) from observation table
        all_labels = []
        for idx, row in enumerate(self.observation_table.items):
            s_region_str = row.get('s_region', '')
            if not s_region_str or (isinstance(s_region_str, str) and s_region_str.strip() == ''):
                continue
            if not is_stcs_string(s_region_str):
                continue
            all_labels.append(idx)

        if not all_labels:
            return

        # Remove footprints from all viewers
        for viewer in image_viewers:
            # Only remove labels that actually exist in this viewer
            existing_labels = [
                mark.label for mark in viewer.figure.marks
                if isinstance(mark, RegionOverlay)
            ]
            labels_to_remove = [label for label in all_labels if label in existing_labels]
            if labels_to_remove:
                viewer._remove_region_overlay(region_label=labels_to_remove)

    def _sync_footprint_selection_to_viewers(self):
        """Sync footprint selection state across all viewers based on table selection."""
        image_viewers = self._get_image_viewers()
        if not image_viewers:
            return

        selected_indices = set()
        for row in self.observation_table.selected_rows:
            idx = self.observation_table.items.index(row)
            selected_indices.add(idx)

        valid_labels = []
        for idx, row in enumerate(self.observation_table.items):
            s_region_str = row.get('s_region', '')
            if not s_region_str or (isinstance(s_region_str, str) and s_region_str.strip() == ''):
                continue
            if not is_stcs_string(s_region_str):
                continue
            valid_labels.append(idx)

        # Update selection in all viewers
        for viewer in image_viewers:
            for label in valid_labels:
                if label in selected_indices:
                    viewer._select_region_overlay(region_label=label)
                else:
                    viewer._deselect_region_overlay(region_label=label)

    def _on_viewer_added(self, msg):
        """When a new viewer is created, add footprints if toolbar is enabled."""
        if not self.custom_toolbar_enabled:
            return
        if not (self.parsed_input_is_query and self.treat_table_as_query):
            return

        viewer = self.app.get_viewer_by_id(msg.viewer_id)
        if viewer is None:
            return
        # Check if it's an image viewer
        if not _is_image_viewer(viewer):
            return
        if 's_region' not in self.observation_table.headers_avail:
            return
        # Check if viewer has valid WCS before attempting to add footprints
        # New viewers may not have data/WCS yet, so we skip and wait for AddDataMessage
        if (not hasattr(viewer.state, 'reference_data') or
            viewer.state.reference_data is None or
            not hasattr(viewer.state.reference_data, 'coords') or
                viewer.state.reference_data.coords is None):
            # No WCS yet - footprints will be added later when data is loaded
            # via _on_viewer_data_added handler
            return
        # Viewer has WCS, safe to add footprints
        self._add_footprints_to_viewer(viewer)

    def _add_footprints_to_viewer(self, viewer):
        # Build regions from table
        regions = []
        region_labels = []
        for idx, row in enumerate(self.observation_table.items):
            s_region_str = row.get('s_region', '')
            if not s_region_str or (isinstance(s_region_str, str) and s_region_str.strip() == ''):
                continue
            if not is_stcs_string(s_region_str):
                continue
            region = stcs_string2region(s_region_str)
            regions.append(region)
            region_labels.append(idx)

        if not regions:
            return
        viewer._add_region_overlay(
            region=regions,
            region_label=region_labels,
            selected=False,
            fill_opacities=[0]
        )
        # Sync selection state with table
        self._sync_footprint_selection_to_viewers()

    def on_file_select_changed(self, _=None):
        self._clear_cache('output')
        self._update_format_items()

    @with_spinner('spinner', 'searching for valid formats...')
    def _update_format_items(self):
        # NOTE: this will call self.output
        self.format._update_items()
        self.target._update_items()  # assumes format._importers is updated from above
        # ensure the importer updates even if the format selection remains fixed
        self._on_format_selected_changed()

    def get_selected_url(self):
        if len(self.file_table.selected_rows) != 1:
            return None
        url = self.file_table.selected_rows[0]['url']
        if not url.startswith(('http://', 'https://', 'mast:', 'ftp:', 's3:')):
            return f'https://mast.stsci.edu/search/{self.guess_mission(url)}/api/v0.1/retrieve_product?product_name={url}'  # noqa
        return url

    @with_spinner('spinner', 'downloading file...')
    def _download_from_file_table(self):
        url = self.get_selected_url().strip()
        if not url:
            return None
        return download_uri_to_path(url,
                                    cache=self.file_cache,
                                    local_path=self.file_local_path,
                                    timeout=self.file_timeout)

    @cached_property
    def output(self):
        if self.parsed_input_is_query and self.treat_table_as_query:
            return self._download_from_file_table()
        else:
            return self.parsed_input

    @property
    def user_api(self):
        return LoaderUserApi(self)

    @property
    def default_label(self):
        # override by subclass to provide a default label to the importer
        # importers can then decide whether to use this or not.
        return None

    @property
    def parser(self):
        # give access to the parser used by the selected importer
        return self.importer._parser

    @property
    def importer(self):
        # give access to the importer defined by the user-selection on format
        if not self.format.selected:
            raise ValueError("must select a format before accessing importer")
        return self.format._importers[self.format.selected]

    def load(self):
        """
        Import into jdaviz with all selected options.
        """
        return self.importer()

    @observe('target_selected')
    def _on_target_selected_changed(self, change={}):
        def matches_target_factory(target):
            def matches_target(importer):
                return importer.target.get('label', '') == target
            return matches_target

        if self._restrict_to_target is not None:
            self.format.filters = [matches_target_factory(self._restrict_to_target)]
        elif self.target_selected == 'Any':
            self.format.filters = []
        else:
            self.format.filters = [matches_target_factory(self.target_selected)]

    def _get_valid_import_formats(self):
        if self._restrict_to_target is not None:
            # if the resolver was initialized with a list of valid import formats,
            # use that instead of the registry
            return [self._restrict_to_target]
        # fallback on all items available in the registry
        return loader_importer_registry.members.keys()

    @observe('format_selected')
    def _on_format_selected_changed(self, change={}):
        if self.format_selected == '':
            self.importer_widget = ''

            if (hasattr(self.format._invalid_importers, 'keys') and
               self.format._invalid_importers.keys()) and self.target_selected:
                # if no valid importer for format_selected, provide supported sources/formats
                # to user in a warning message
                self.valid_import_formats = ", ".join(self._get_valid_import_formats())
        else:
            self.importer_widget = "IPY_MODEL_" + self.importer.model_id
            self.valid_import_formats = ''

            self.importer.reset_and_check_existing_data_in_dc()

    def close_in_tray(self, close_sidebar=False):
        """
        Close the loader in the sidebar/tray.

        Parameters
        ----------
        close_sidebar : bool
            Whether to also close the sidebar itself.
        """
        if self.close_callback is not None:
            self.close_callback()
        if close_sidebar:
            self.app.state.drawer_content = ''

    def open_in_tray(self):
        """
        Show this resolver in the sidebar tray.
        """
        if self.set_active_loader_callback is None:
            raise NotImplementedError("set_active_loader_callback must be set to open dialog to specific tab")  # noqa
        self.set_active_loader_callback(self._registry_label)
        if self.open_callback is not None:
            self.open_callback()


class BaseConeSearchResolver(BaseResolver):
    viewer_items = List([]).tag(sync=True)
    viewer_selected = Unicode().tag(sync=True)

    source = Unicode("").tag(sync=True)
    coord_follow_viewer_pan = Bool(False).tag(sync=True)
    viewer_centered = Bool(False).tag(sync=True)
    coordframe_choices = List([]).tag(sync=True)
    coordframe_selected = Unicode("icrs").tag(sync=True)
    radius = Float(1).tag(sync=True)
    radius_unit_items = List().tag(sync=True)
    radius_unit_selected = Unicode("deg").tag(sync=True)

    max_results = IntHandleEmpty(1000).tag(sync=True)
    returned_no_results = Bool(False).tag(sync=True)
    returned_max_results = Bool(False).tag(sync=True)

    results_loading = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._output = None

        self.viewer = ViewerSelect(
            self, "viewer_items", "viewer_selected", manual_options=["Manual"],
            filters=['is_image_viewer']
        )

        self.coordframe = SelectPluginComponent(
            self, items="coordframe_choices", selected="coordframe_selected"
        )
        self.coordframe.choices = [frame.lower() for frame in all_astropy_frames]
        self.coordframe.selected = self.coordframe.choices[0]

        self.radius_unit = UnitSelectPluginComponent(
            self, items="radius_unit_items", selected="radius_unit_selected"
        )
        self.radius_unit.choices = ["deg", "rad", "arcmin", "arcsec"]
        self.radius_unit.selected = "deg"

        self.hub.subscribe(self, AddDataMessage, handler=self.vue_center_on_data)
        self.hub.subscribe(self, RemoveDataMessage, handler=self.vue_center_on_data)
        self.hub.subscribe(self, LinkUpdatedMessage, handler=self.vue_center_on_data)

    @observe("viewer_selected", type="change")
    def vue_viewer_changed(self, _=None):
        # Check mixin object initialized
        if not hasattr(self, "viewer"):
            return

        # Clear all existing subscriptions and resubscribe to selected viewer
        # NOTE: Viewer subscription needed regardless of coord_follow_viewer_pan in order
        #   to detect when coords are centered on viewer, regardless of viewer tracking
        for viewer in self.viewer.viewers:
            if viewer == self.viewer.selected_obj:
                viewer.state.add_callback(
                    "zoom_center_x",
                    lambda callback: self.vue_center_on_data(user_zoom_trigger=True),
                )
                viewer.state.add_callback(
                    "zoom_center_y",
                    lambda callback: self.vue_center_on_data(user_zoom_trigger=True),
                )
            else:
                # If not subscribed anyways, remove_callback should produce a no-op
                viewer.state.remove_callback(
                    "zoom_center_x",
                    lambda callback: self.vue_center_on_data(user_zoom_trigger=True),
                )
                viewer.state.remove_callback(
                    "zoom_center_y",
                    lambda callback: self.vue_center_on_data(user_zoom_trigger=True),
                )
        self.vue_center_on_data()

    @observe("coord_follow_viewer_pan", type="change")
    def _toggle_viewer_pan_tracking(self, _=None):
        """Detects when live viewer tracking toggle is clicked and centers on data if necessary"""
        # Center on data if we're enabling the toggle
        if self.coord_follow_viewer_pan:
            self.vue_center_on_data()

    def vue_center_on_data(self, _=None, user_zoom_trigger=False):
        """
        This vue method serves two purposes:
        * UI entrypoint for the manual viewer center button
        * Callback method for user panning (sub'ed to zoom_center_x/zoom_center_y)
        """
        # If plugin is in "Manual" mode, we should never
        # autocenter and potentially wipe the user's data
        if not self.viewer_selected or self.viewer_selected == "Manual":
            return

        # If the user panned but tracking not enabled, don't recenter
        if (user_zoom_trigger) and not self.coord_follow_viewer_pan:
            # Thus, we're no longer centered
            self.viewer_centered = False
            return

        self.center_on_data()

    def center_on_data(self):
        """
        If data is present in the default viewer, center the plugin's coordinates on
        the viewer's center WCS coordinates.
        """
        if not hasattr(self, "viewer"):
            # mixin object not yet initialized
            return

        # gets the current viewer
        viewer = self.viewer.selected_obj

        # nothing happens in the case there is no image in the viewer
        # additionally if the data does not have WCS
        if (
            len(self.app._jdaviz_helper.data_labels) < 1
            or viewer.state.reference_data is None
            or viewer.state.reference_data.coords is None
        ):
            self.source = ""
            return

        # Obtain center point of the current image and convert into sky coordinates
        if self.app._jdaviz_helper.plugins["Orientation"].align_by == "WCS":
            skycoord_center = SkyCoord(
                viewer.state.zoom_center_x, viewer.state.zoom_center_y, unit="deg"
            )
        else:
            skycoord_center = viewer.state.reference_data.coords.pixel_to_world(
                viewer.state.zoom_center_x, viewer.state.zoom_center_y
            )

        # Extract SkyCoord values as strings for plugin display
        ra_deg = skycoord_center.ra.deg
        dec_deg = skycoord_center.dec.deg
        frame = skycoord_center.frame.name.lower()

        # Show center value in plugin
        self.source = f"{ra_deg} {dec_deg}"
        self.coordframe_selected = frame

        self.viewer_centered = True

    @property
    def is_valid(self):
        # these resolvers do not accept any direct, (default_input = None), so can
        # always be considered valid
        return True


def find_matching_resolver(app, inp=None, resolver=None, format=None, target=None, **kwargs):
    formats = format if isinstance(format, (list, tuple)) else [format]
    invalid_resolvers = {}
    valid_resolvers = []
    for resolver_name, Resolver in loader_resolver_registry.members.items():
        if resolver_name == 'file drop':
            # no API input, so let's avoid always returning the confusing
            # message that default_input is undefined
            continue
        if resolver is not None and resolver != resolver_name:
            invalid_resolvers[resolver_name] = f'not {resolver}'
            continue
        try:
            this_resolver = Resolver.from_input(app, inp, format=format, **kwargs)
        except Exception as e:  # nosec
            invalid_resolvers[resolver_name] = f'resolver exception: {e}'
            if resolver_name == 'url' and 'timeout' in str(e):
                raise e
            continue
        try:
            is_valid = this_resolver.is_valid
        except Exception as e:  # nosec
            invalid_resolvers[resolver_name] = f'is_valid exception: {e}'
            is_valid = False
        if not is_valid:
            invalid_resolvers.setdefault(resolver_name, 'not valid')
            continue

        if target is not None:
            try:
                this_resolver.target = target
            except ValueError:
                invalid_resolvers[resolver_name] = this_resolver.format._invalid_importers
                continue
        if not len(this_resolver.format.items):
            invalid_resolvers[resolver_name] = this_resolver.format._invalid_importers
            continue

        for fmt_item in this_resolver.format.items:
            if (format is not None
                and not any([format in (fmt_item['label'],
                                        fmt_item['parser'],
                                        fmt_item['importer'])
                             for format in formats])):
                invalid_resolvers[resolver_name] = this_resolver.format._invalid_importers
                continue
            this_resolver.format.selected = fmt_item['label']
            valid_resolvers.append((this_resolver, resolver_name, fmt_item['label']))

    if len(valid_resolvers) == 0:
        raise ValueError("no valid loaders found for input, tried", invalid_resolvers)
    elif len(valid_resolvers) > 1:
        vrs = [f"resolver={vr[1]} > format={vr[2]}" for vr in valid_resolvers]
        raise ValueError(f"multiple valid loaders found for input: {vrs}")
    else:
        return valid_resolvers[0][0].user_api



import os
import re

import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDData
from glue.core.roi import CircularROI, EllipticalROI
from regions import Regions, CircleSkyRegion
from specutils import Spectrum, SpectralRegion
from pathlib import Path
from astropy.wcs import WCS


@pytest.mark.usefixtures('_jail')
class TestExportSubsets:
    """
    Tests for exporting subsets. Currently limited to non-composite spatial
    subsets.
    """

    def test_basic_export_subsets_imviz(self, imviz_helper):

        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)
        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=250, yc=250, radius=100))

        export_plugin = imviz_helper.plugins['Export']._obj

        # Check initialization of these two
        assert export_plugin.subset_invalid_msg == ''
        assert export_plugin.subset_format_invalid_msg == ''

        export_plugin.subset.selected = 'Subset 1'

        # Make no assumptions on default since the 'default' is now
        # set by the first option in the list of available formats
        # (but it's *almost* 100% likely to be fits)
        spatial_valid_formats = ['fits', 'reg']
        assert export_plugin.subset_format.selected in spatial_valid_formats

        for current_format in spatial_valid_formats:
            export_plugin.subset_format.selected = current_format
            assert export_plugin.subset_format.selected == current_format
            assert export_plugin.subset_invalid_msg == ''  # for non-composite spatial
            assert export_plugin.subset_format_invalid_msg == ''

            assert export_plugin.filename.value.endswith(f".{current_format}")
            export_plugin.export()
            assert os.path.isfile(export_plugin.filename.value)

            # changing file name and catching the result (the new filename/path)
            # for checking the file
            new_filename = 'test'
            export_plugin.filename.value = new_filename
            output_filename = export_plugin.export()
            assert os.path.isfile(f'{new_filename}.{current_format}')
            assert os.path.isfile(output_filename)

            if current_format == 'fits':
                # read exported file back in
                with fits.open(output_filename) as hdu:
                    fits_region = hdu[1].data[0]

                assert fits_region[0] == 'circle'
                assert fits_region[1] == fits_region[2] == 250.0
                assert fits_region[3] == 100.0
                assert fits_region[4] == 0.0

            elif current_format == 'reg':
                # read exported file back in
                region = Regions.read(output_filename)[0]
                assert region.center.x == 250.0
                assert region.center.y == 250.0
                assert region.radius == 100.0

        # test that invalid file extension raises an error
        with pytest.raises(ValueError,
                           match=re.escape("'x' not one of ['fits', 'reg', 'ecsv'], reverting selection to 'reg'")):  # noqa
            export_plugin.subset_format.selected = 'x'

    def test_not_implemented(self, cubeviz_helper, spectral_cube_wcs):
        """
        Test that trying to export non-supported subsets
        (spectral and composite) produces
        the correct warning message to display in UI).
        """

        data = Spectrum(flux=np.ones((500, 500, 2)) * u.nJy,
                        wcs=spectral_cube_wcs)
        cubeviz_helper.load_data(data)
        subset_plugin = cubeviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=255, yc=255, radius=50))
        subset_plugin.import_region(CircularROI(xc=200, yc=250, radius=50), edit_subset='Subset 1',
                                    combination_mode='and')

        export_plugin = cubeviz_helper.plugins['Export']._obj
        export_plugin.subset.selected = 'Subset 1'

        assert export_plugin.subset_invalid_msg == 'Export for composite subsets not yet supported.'

    def test_export_subsets_wcs(self, imviz_helper, spectral_cube_wcs):

        # using cube WCS instead of 2d imaging wcs for consistency with
        # cubeviz test. accessing just the spatial part of this.
        wcs = spectral_cube_wcs.celestial

        data = NDData(np.ones((500, 500)) * u.nJy, wcs=wcs)

        imviz_helper.load_data(data)  # load data twice so we can link them
        imviz_helper.load_data(data)

        imviz_helper.link_data(align_by='wcs')
        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=8, yc=6, radius=.2))

        export_plugin = imviz_helper.plugins['Export']._obj
        export_plugin.subset.selected = 'Subset 1'

        assert export_plugin.subset_invalid_msg == ''  # for non-composite spatial

        # test changing link type results in an output file with a sky region
        export_plugin.subset_format.selected = 'reg'
        export_plugin.filename_value = 'sky_region.reg'
        export_plugin.export()

        assert os.path.isfile('sky_region.reg')

        assert isinstance(Regions.read('sky_region.reg')[0], CircleSkyRegion)

    def test_basic_export_subsets_cubeviz(self, cubeviz_helper, spectral_cube_wcs):

        data = Spectrum(flux=np.ones((128, 128, 256)) * u.nJy, wcs=spectral_cube_wcs)

        # Subset 1, Spatial Subset
        cubeviz_helper.load_data(data)
        subset_plugin = cubeviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=50, yc=50, radius=10))

        # Subset 2, Spectral Subset
        spectral_axis_unit = u.Unit(
            cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
        subset_plugin.import_region(SpectralRegion(5 * spectral_axis_unit,
                                                   15.5 * spectral_axis_unit))

        export_plugin = cubeviz_helper.plugins['Export']._obj

        # Assert that the default name is set and that they're in order
        # Mostly to confirm the use of spatial/specctral_subset variables
        assert export_plugin.subset.choices[0] == 'Subset 1'
        spatial_subset = 'Subset 1'

        assert export_plugin.subset.choices[1] == 'Subset 2'
        spectral_subset = 'Subset 2'

        # No subset is selected by default
        assert export_plugin.subset.selected == ''

        export_plugin.subset.selected = spatial_subset
        spatial_valid_formats = ['fits', 'reg']
        # Assert that the first subset created has the first format available (fits).
        assert export_plugin.subset_format.selected in spatial_valid_formats

        for current_format in spatial_valid_formats:
            # Check that the format and filename are set correctly
            export_plugin.subset_format.selected = current_format
            assert export_plugin.subset_format.selected == current_format
            assert export_plugin.filename.value.endswith(current_format)

            assert export_plugin.subset_invalid_msg == ''  # for non-composite spatial
            assert export_plugin.subset_format_invalid_msg == ''

            # Attempt export
            export_plugin.export()
            assert os.path.isfile(export_plugin.filename.value)

            # changing file name and catching the result (the new filename/path)
            # for checking the file
            new_filename = 'test'
            export_plugin.filename.value = new_filename
            output_filename = export_plugin.export()
            assert os.path.isfile(f'{new_filename}.{current_format}')
            assert os.path.isfile(output_filename)

            if current_format == 'fits':
                # read exported file back in
                with fits.open(output_filename) as hdu:
                    fits_region = hdu[1].data[0]

                assert fits_region[0] == 'circle'
                assert fits_region[1] == fits_region[2] == 50.0
                assert fits_region[3] == 10.0
                assert fits_region[4] == 0.0

            elif current_format == 'reg':
                # read exported file back in
                region = Regions.read(output_filename)[0]
                assert region.center.x == 50.0
                assert region.center.y == 50.0
                assert region.radius == 10.0

            # Finally, attempt to export with Spectral subset
            export_plugin.subset.selected = spectral_subset
            # subset_format_invalid_msg should already be populated since
            # the selected format is fits/reg
            assert export_plugin.subset_format_invalid_msg != ''
            with pytest.raises(ValueError,
                               match=f"Export of '{spectral_subset}' "
                                       f"in '{current_format}' format is not supported."):  # noqa
                export_plugin.export()

            # Reset to spatial
            export_plugin.subset.selected = spatial_subset

        # Overwrite not enable, so no-op with warning.
        export_plugin.export(raise_error_for_overwrite=False)
        assert export_plugin.overwrite_warn

        # Changing filename should clear warning.
        old_filename = export_plugin.filename_value
        export_plugin.filename_value = "foo"
        assert not export_plugin.overwrite_warn
        export_plugin.filename_value = old_filename

        # Overwrite not enable, but with exception from API by default.
        with pytest.raises(FileExistsError, match=".* exists but overwrite=False"):
            export_plugin.export()
        assert export_plugin.overwrite_warn

        # User forces overwrite.
        export_plugin.export(overwrite=True)
        assert not export_plugin.overwrite_warn

        # test that invalid file extension raises an error
        with pytest.raises(ValueError,
                           match=re.escape(
                               "'x' not one of ['fits', 'reg', 'ecsv'], reverting selection to 'reg'")):  # noqa
            export_plugin.subset_format.selected = 'x'

        # Test that attempting to export with disabled option raises an error
        export_plugin.subset_format.selected = 'ecsv'
        with pytest.raises(ValueError,
                           match=f"Export of '{spatial_subset}' "
                                 f"in 'ecsv' format is not supported."):  # noqa
            export_plugin.export()

        # test that attempting to save a composite subset raises an error
        subset_plugin.import_region(CircularROI(xc=25, yc=25, radius=5), edit_subset='Subset 1',
                                    combination_mode='and')
        subset_plugin.import_region(CircularROI(xc=20, yc=25, radius=5), edit_subset='Subset 1',
                                    combination_mode='and')

        export_plugin.subset.selected = spatial_subset
        # A user *must* (re)select the subset for the subset_invalid_msg to be filled
        # if the subset is already selected and subsequently edited, the message
        # will be an empty string
        assert export_plugin.subset_invalid_msg == 'Export for composite subsets not yet supported.'
        # However it *will* still fail on export.
        with pytest.raises(NotImplementedError,
                           match='Subset cannot be exported - Export for composite subsets not yet supported.'):  # noqa
            export_plugin.export()

        export_plugin.subset.selected = spectral_subset
        export_plugin.filename_value = "test_spectral_region"
        export_plugin.export()
        assert os.path.isfile('test_spectral_region.ecsv')

        # Revert spectral subset format selection to fits and check
        # that the format message is populated
        export_plugin.subset_format.selected = 'fits'
        assert export_plugin.subset_format_invalid_msg != ''
        # Delete the spectral subset and check that the message is reset
        cubeviz_helper.app.delete_subsets(spectral_subset)
        assert export_plugin.subset_format_invalid_msg == ''

    def test_export_stcs_circle_ellipse(self, imviz_helper):
        wcs = WCS({'CTYPE1': 'RA---TAN', 'CUNIT1': 'deg', 'CDELT1': -0.0002777777778,
                   'CRPIX1': 1, 'CRVAL1': 9.423508457380343,
                   'CTYPE2': 'DEC--TAN', 'CUNIT2': 'deg', 'CDELT2': 0.0002777777778,
                   'CRPIX2': 1, 'CRVAL2': -33.71313112382379})
        arr = np.arange(40000).reshape(200, 200)
        ndd = NDData(arr, wcs=wcs)
        imviz_helper.load_data(ndd)
        imviz_helper.link_data(align_by='wcs')

        subset_plugin = imviz_helper.plugins['Subset Tools']
        export_plugin = imviz_helper.plugins['Export']._obj

        subset_plugin.import_region(CircularROI(xc=40, yc=50, radius=10))
        export_plugin.subset.selected = 'Subset 1'
        export_plugin.subset_format.selected = 'stcs'
        export_plugin.filename_value = 'Subset_1.stcs'
        export_plugin.export()
        with open('Subset_1.stcs') as f:
            contents = f.read().split()
        assert contents[0] == 'CIRCLE'
        assert contents[1] == 'ICRS'
        assert contents[2] == '9.157221'
        assert contents[3] == '-33.435070'
        assert contents[4] == '0.055554'

        subset_plugin.import_region(EllipticalROI(xc=45, yc=55, radius_x=8, radius_y=4, theta=0.2))
        export_plugin.subset.selected = 'Subset 2'
        export_plugin.subset_format.selected = 'stcs'
        export_plugin.filename_value = 'Subset_2.stcs'
        export_plugin.export(overwrite=True)
        with open('Subset_2.stcs') as f:
            contents = f.read().split()
        assert contents[0] == 'ELLIPSE'
        assert contents[1] == 'ICRS'
        assert contents[2] == '9.124032'
        assert contents[3] == '-33.407217'
        assert contents[4] == '0.088886'
        assert contents[5] == '0.044443'
        assert contents[6] == '11.625269'


@pytest.mark.usefixtures('_jail')
def test_export_cubeviz_spectrum_viewer(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')

    ep = cubeviz_helper.plugins["Export"]
    ep.viewer = 'spectrum-viewer'
    ep.viewer_format = 'png'
    ep.export()

    ep.viewer_format = 'svg'
    ep.export()


@pytest.mark.usefixtures('_jail')
def test_export_data(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')
    mm = cubeviz_helper.plugins["Moment Maps"]
    mm.dataset = 'test[FLUX]'
    mm._obj.dataset_selected = 'test[FLUX]'
    mm.n_moment = 0
    mm.calculate_moment()
    assert mm._obj.results_label == 'moment 0'
    cubeviz_helper.app.add_data_to_viewer(
        cubeviz_helper._default_flux_viewer_reference_name, 'moment 0'
    )
    ep = cubeviz_helper.plugins["Export"]._obj
    assert 'test[FLUX]' not in ep.dataset.choices

    ep.dataset_selected = 'moment 0'
    assert ep.dataset_format.selected == 'fits'
    ep.export()
    assert os.path.isfile(ep.filename.value)
    assert ep.data_invalid_msg == ''


def test_disable_export_for_unsupported_units(specviz2d_helper):
    dn_per_s = u.DN / u.s
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    data = Spectrum(flux=data*dn_per_s, spectral_axis=data[3]*u.um)
    specviz2d_helper.load_data(data)

    gs = specviz2d_helper.plugins["Gaussian Smooth"]
    smooth_source_dataset = "Spectrum 1D"
    gs.dataset = smooth_source_dataset
    gs.stddev = 3
    gs.smooth(add_data=True)

    ep = specviz2d_helper.plugins["Export"]._obj
    assert "Spectrum 1D smooth stddev-3.0" in ep.dataset.choices
    ep.dataset_selected = "Spectrum 1D smooth stddev-3.0"
    assert ep.dataset.selected_obj.unit == "DN/s"
    assert ep.data_invalid_msg == "Export Disabled: The unit DN / s could not be saved in native FITS format."  # noqa


class TestExportPluginPlots:

    def test_basic_export_plugin_plots(self, imviz_helper):
        """
        Test basic funcionality of exporting plugin plots
        from the export plugin. Tests on the 'Plot Options: stretch_hist'
        plot, which exists upon loading data, and also that plots that
        may have been initialized but are empty are not displayed in
        the Export plugin.
        """
        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)

        export_plugin = imviz_helper.plugins['Export']._obj
        export_plugin.plugin_plot.selected = 'Plot Options: stretch_hist'

        assert export_plugin.plugin_plot_format.selected == 'png'  # should be default format
        # and change file type
        export_plugin.plugin_plot_format.selected = 'svg'

        assert export_plugin.filename.value.endswith('.svg')
        # change filename
        export_plugin.filename_value = 'test_export_plugin_plot'

        # just check that it doesn't crash, since we can't download
        export_plugin.export()

        # make sure that the only valid option for export is this plugin,
        # not the other plots that exist but are empty (ap phot and line profile)
        # this might change down the line if new plots are added.
        available_plots = [x['label'] for x in export_plugin.plugin_plot.items]
        assert len(available_plots) == 1
        assert available_plots[0] == 'Plot Options: stretch_hist'

    def test_ap_phot_plot_export(self, imviz_helper):
        """
        Test export functionality for plot from the aperture photometry
        plugin.
        """

        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)

        export_plugin = imviz_helper.plugins['Export']._obj
        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=250, yc=250, radius=100))

        phot_plugin = imviz_helper.app.get_tray_item_from_name('imviz-aper-phot-simple')
        phot_plugin.aperture_selected = 'Subset 1'

        phot_plugin.vue_do_aper_phot()
        assert phot_plugin.plot_available

        available_plots = [x['label'] for x in export_plugin.plugin_plot.items]
        assert 'Aperture Photometry: plot' in available_plots

        export_plugin.plugin_plot.selected = 'Aperture Photometry: plot'

        # change filename
        export_plugin.filename_value = 'test_export_plugin_plot'
        # and change file type
        export_plugin.plugin_plot_format.selected = 'svg'

        # just check that it doesn't crash, since we can't download
        export_plugin.export()

    def test_figure_export(self, imviz_helper):

        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)

        export_plugin = imviz_helper.plugins['Export']._obj

        export_plugin.export(filename=None)

        # attempt to save a figure back to back
        try:
            export_plugin.export(filename='img.png')
        except ValueError as e:
            assert str(e) == "previous png export is still in progress. Wait to complete before making another call to save_figure"  # noqa: E501

    def test_filepath_convention(self, imviz_helper):
        data = NDData(np.ones((500, 500)) * u.nJy)
        imviz_helper.load_data(data)
        export_plugin = imviz_helper.plugins['Export']._obj

        # Set filename value using OS-independent Path methods
        export_plugin.filename_value = str(Path('/') / 'img.png')
        assert os.path.abspath(export_plugin.filepath) == os.path.abspath(export_plugin.filename_value)  # noqa: E501

        export_plugin.filename_value = str(Path('~') / 'img.png')
        expected_path = str(Path('~').expanduser() / 'img.png')
        assert export_plugin.filepath == expected_path

        export_plugin.filename_value = str(Path('..') / 'img.png')
        expected_path = str((Path('..') / 'img.png').resolve())
        assert export_plugin.filepath == expected_path

        export_plugin.filename_value = str(Path('.') / 'img.png')
        expected_path = str((Path('.') / 'img.png').resolve())
        assert export_plugin.filepath == expected_path



import itertools
import numpy as np
import warnings

from astropy.nddata import StdDevUncertainty
from specutils import Spectrum, SpectrumList, SpectrumCollection
from traitlets import List, Bool, Any, observe

from jdaviz.core.unit_conversion_utils import (to_flux_density_unit,
                                               spectrum_ensure_flux_density_unit)
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import (BaseImporterToDataCollection,
                                           _spectrum_assign_component_type)
from jdaviz.core.template_mixin import SelectFileExtensionComponent
from jdaviz.core.user_api import ImporterUserApi
from jdaviz.core.events import SnackbarMessage
from jdaviz.utils import create_data_hash


__all__ = ['SpectrumListImporter', 'SpectrumListConcatenatedImporter']


@loader_importer_registry('1D Spectrum List')
class SpectrumListImporter(BaseImporterToDataCollection):
    template_file = __file__, "spectrum_list.vue"

    sources_items = List().tag(sync=True)
    sources_selected = Any().tag(sync=True)
    sources_multiselect = Bool(True).tag(sync=True)

    input_in_sb = Bool(False).tag(sync=True)
    convert_to_flux_density = Bool(True).tag(sync=True)

    disable_dropdown = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver
        elif self.app.config == 'specviz':
            self.data_label_default = '1D Spectrum'
        else:
            self.data_label_default = '1D Spectrum'

        sources_options = []

        if isinstance(self.input, Spectrum):
            speclist_input = SpectrumList(self.input_to_list_of_spec(self.input))
        else:
            speclist_input = self.input

        for index, spec in enumerate(speclist_input):
            if self.is_wfssmulti(spec):
                # ver, name are stand-ins for exposure and source_id
                # ver == exposure, name == source_id
                ver, name = self._extract_exposure_sourceid(spec)
                exposure_label = f"Exposure {ver}"

                label = f"{exposure_label}, Source ID: {name}"
                # Flipping the two from the variable naming convention
                name_ver = f"{ver}_{name}"
                suffix = f"EXP-{ver}_ID-{name}"

            else:
                name_ver = index
                name = index
                ver = index
                label = f"1D Spectrum at index: {index}"
                suffix = f"index-{index}"

            spec = self._apply_spectral_mask(spec)
            sources_options.append({'label': label,
                                    'index': index,
                                    'name': str(name),
                                    'ver': str(ver),
                                    'name_ver': str(name_ver),
                                    'suffix': suffix,
                                    'data_hash': create_data_hash(spec),
                                    'obj': spec})

        self.sources = SelectFileExtensionComponent(self,
                                                    items='sources_items',
                                                    selected='sources_selected',
                                                    multiselect='sources_multiselect',
                                                    manual_options=sources_options)

        self.sources.selected = [self.sources.choices[0]]
        self.data_hashes = self.sources.data_hashes
        self.hash_map_to_label = dict(zip(self.sources.data_hashes, self.sources.labels))

        # TODO: This observer will likely be removed in follow-up effort
        # If the resolver format is set to "1D Spectrum List", then we
        # only enable the import button if at least one spectrum is selected.
        self.resolver.observe(self._on_format_selected_change, names='format_selected')

    def _apply_kwargs(self, kwargs):
        applied_kwargs = super()._apply_kwargs(kwargs)
        if 'sources' not in applied_kwargs:
            msg_str = (f"The default source selection ({self.sources.selected}) will be used.\n"
                       f"To load additional sources, please specify them via dropdown or "
                       f"as follows:\n'{self.config}.load(filename, sources = [...]).")
            msg = SnackbarMessage(msg_str, color='warning', sender=self, timeout=10000)
            self.app.hub.broadcast(msg)
            warnings.warn(msg_str)
        return applied_kwargs

    @staticmethod
    def _get_supported_viewers():
        return [{'label': '1D Spectrum', 'reference': 'spectrum-1d-viewer'}]

    @property
    def user_api(self):
        expose = ['sources', 'convert_to_flux_density']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz'):
            # NOTE: temporary during deconfig process
            return False
        # TODO: should this be split into two loaders?
        # should a loader take a single input type, output a single output type,
        # or just have a consistent data_label and viewer?

        # If the input is a SpectrumList or SpectrumCollection, it
        # must be non-empty.
        if isinstance(self.input, (SpectrumList, SpectrumCollection)):
            return len(self.input) > 0

        return self._is_2d_spectrum

    @observe('sources_selected')
    def _on_sources_selected_change(self, change={}):
        if len(self.sources_selected) == 0:
            self.import_disabled = True
        else:
            self.import_disabled = False

            self.input_in_sb = bool(np.any([sp.flux.unit.physical_type == 'surface brightness'
                                            for sp in self.sources.selected_obj]))

    def _on_format_selected_change(self, change={}):
        if change['new'] == '1D Spectrum List':
            # Only perform this check if the selected importer is SpectrumListImporter
            # Otherwise other valid importers (in the case of Spectrum2D)
            # will also run and reset import_disabled
            if 'SpectrumListImporter' in str(type(self)).split('.')[-1]:
                self._on_sources_selected_change()

        elif change['new'] == '1D Spectrum Concatenated':
            # 2D Spectra load all for concatenated
            if self._is_2d_spectrum:
                self.import_disabled = False
            else:
                self._on_sources_selected_change()

        else:
            self.import_disabled = False

    def input_to_list_of_spec(self, inp):

        def this_row(field, i):
            if field is None:
                return None
            return field[i, :]

        if isinstance(inp, Spectrum):
            if inp.flux.ndim == 1:
                return [self._apply_spectral_mask(inp)]

            return [Spectrum(spectral_axis=inp.spectral_axis,
                             flux=this_row(inp.flux, i),
                             uncertainty=this_row(inp.uncertainty, i),
                             mask=this_row(inp.mask, i),
                             meta=inp.meta)
                    for i in range(inp.flux.shape[0])]

        elif isinstance(inp, (SpectrumList, SpectrumCollection)):
            return itertools.chain(*[self.input_to_list_of_spec(spec) for spec in inp])

        else:
            raise NotImplementedError(f"{inp} is not supported")

    @property
    def output(self):
        if not self.is_valid:  # pragma: nocover
            return None
        if self.input_in_sb and self.convert_to_flux_density:
            return [spectrum_ensure_flux_density_unit(sp) for sp in self.sources.selected_obj]
        else:
            return self.sources.selected_obj

    @staticmethod
    def is_wfssmulti(spec):
        return 'WFSSMulti' in spec.meta.get('header', {}).get('DATAMODL', '')

    @staticmethod
    def _extract_exposure_sourceid(spec):
        """
        Generate a label for WFSSMulti sources based on the header information.
        """
        header = spec.meta.get('header', {})
        exp_num = header.get('EXPGRPID', '0_0_0').split('_')[-2]
        source_id = str(spec.meta.get('source_id', ''))

        return exp_num, source_id

    @staticmethod
    def _has_mask(spec):
        if hasattr(spec, 'mask'):
            if spec.mask is not None and len(spec.mask):
                return True
        return False

    def _is_fully_masked(self, spec):
        if self._has_mask(spec):
            # all == True implies the entire array is masked and unusable
            if all(spec.mask):
                return True
        return False

    @property
    def _is_2d_spectrum(self):
        return isinstance(self.input, Spectrum) and self.input.flux.ndim == 2

    def _apply_spectral_mask(self, spec):
        # The masks (spec.spectral_axis.mask and spec.mask) for WFSS L3 spectra
        # may not be equivalent, so we only apply the spectral_axis mask to avoid
        # a Specutils error. Specutils expects the spectral axis to be strictly
        # increasing/decreasing so applying the 'full' mask may throw that error.
        if self._has_mask(spec.spectral_axis) and not self._is_fully_masked(spec):
            mask = spec.spectral_axis.mask
            # NOTE: Something breaks when the following is attempted instead
            # of the current implementation
            #
            # spec.mask = spec.mask | spec.spectral_axis.mask
            # return spec
            return Spectrum(
                spectral_axis=spec.spectral_axis[~mask],
                flux=spec.flux[~mask],
                uncertainty=spec.uncertainty[~mask],
                mask=mask[~mask],
                meta=spec.meta)

        return spec

    def assign_component_type(self, comp_id, comp, units, physical_type):
        return _spectrum_assign_component_type(comp_id, comp, units, physical_type)

    def __call__(self):
        if not self.sources.selected:
            raise ValueError("No sources selected.")

        with self.app._jdaviz_helper.batch_load():
            for spec_obj, item_dict in zip(self.output, self.sources.selected_item_list):
                data_label = f"{self.data_label_value}_{item_dict['suffix']}"
                self.add_to_data_collection(spec_obj, data_label,
                                            data_hash=item_dict.get('data_hash'))


def combine_lists_to_1d_spectrum(wl, fnu, dfnu, wave_units, flux_units):
    """
    Combine lists of 1D spectra into a composite `~specutils.Spectrum` object.

    Parameters
    ----------
    wl : list of `~astropy.units.Quantity`s
        Wavelength in each spectral channel
    fnu : list of `~astropy.units.Quantity`s
        Flux in each spectral channel
    dfnu : list of `~astropy.units.Quantity`s or None
        Uncertainty on each flux

    Returns
    -------
    spec : `~specutils.Spectrum`
        Composite 1D spectrum.
    """
    # COPIED FROM specviz.plugins.parsers since cannot import
    wlallarr = np.array(wl)
    fnuallarr = np.array(fnu)
    srtind = np.argsort(wlallarr)
    if dfnu is not None:
        dfnuallarr = np.array(dfnu)
        fnuallerr = dfnuallarr[srtind]
    wlall = wlallarr[srtind]
    fnuall = fnuallarr[srtind]

    # units are not being handled properly yet.
    if dfnu is not None:
        unc = StdDevUncertainty(fnuallerr * flux_units)
    else:
        unc = None

    spec = Spectrum(flux=fnuall * flux_units, spectral_axis=wlall * wave_units,
                    uncertainty=unc)
    return spec


@loader_importer_registry('1D Spectrum Concatenated')
class SpectrumListConcatenatedImporter(SpectrumListImporter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.select_all_for_concatenation = False

        if self._is_2d_spectrum:
            self.disable_dropdown = True
            # If we select_all() here, then if the user switches back to Spectrum List
            # all items will be selected, which is not the desired behavior.
            self.select_all_for_concatenation = True
            # Enable the import button upon initialization because otherwise having
            # no sources selected will disable it for other valid importers (e.g. Image).
            self.import_disabled = False

    @property
    def output(self):
        if self.select_all_for_concatenation:
            self.sources.select_all()

        spectrum_list = self.sources.selected
        if len(spectrum_list) == 0:
            return []

        # Vectorized collection of all wavelengths, fluxes, and uncertainties
        wl_list = []
        fnu_list = []
        dfnu_list = []
        for spec in self.sources.selected_obj:
            wl = spec.spectral_axis.value
            fnu = spec.flux.value

            if spec.uncertainty is not None:
                dfnu = spec.uncertainty.array
            else:
                dfnu = np.full_like(fnu, np.nan)

            wl_list.append(wl)
            fnu_list.append(fnu)
            dfnu_list.append(dfnu)

        wlallorig = np.concatenate(wl_list)
        fnuallorig = np.concatenate(fnu_list)
        dfnuallorig = np.concatenate(dfnu_list)

        wave_units = spec.spectral_axis.unit
        pixar_sr = getattr(spec, 'meta', {}).get('PIXAR_SR', 1.0)
        flux_units = to_flux_density_unit(spec.flux.unit, pixar_sr)

        return combine_lists_to_1d_spectrum(wlallorig,
                                            fnuallorig,
                                            dfnuallorig,
                                            wave_units,
                                            flux_units)

    def __call__(self):
        data_label = self.data_label_value
        self.add_to_data_collection(self.output, f"{data_label}")

        # Do we need to reset in case user switches back to Spectrum List?
        # self.sources.selected = []
